# -*- coding: utf-8 -*-
"""
Created on April 23 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007

Adapted from @agiovann and @epnev
"""

from __future__ import division
import sys
from PyQt5 import QtCore, QtGui, QtWidgets
import json

import numpy as np
from matplotlib.widgets import Slider
from caiman.utils import visualization

import os
import pickle
from glob import glob
from functools import partial
import traceback
from time import time, sleep
import logging
import caiman as cm
from caiman.source_extraction import cnmf
from caiman.utils.utils import load_dict_from_hdf5
from caiman.source_extraction.cnmf.params import CNMFParams

import cv2
try:
    cv2.setNumThreads(0)
except:
    pass


if not sys.argv[0] == __file__:
    from ..roi_manager import ModuleGUI
    from ...core import ViewerUtils, ViewerWorkEnv
    from ....pyqtgraphCore.widgets.MatplotlibWidget import MatplotlibWidget


def run(batch_dir: str, UUID: str):
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG,
                        format="%(relativeCreated)12d [%(filename)s:%(funcName)20s():%(lineno)s] [%(process)d] %(message)s")

    start_time = time()

    output = {'status': 0, 'output_info': ''}
    n_processes = os.environ['_MESMERIZE_N_THREADS']
    n_processes = int(n_processes)
    file_path = os.path.join(batch_dir, UUID)

    filename = [file_path + '_input.tiff']
    input_params = pickle.load(open(file_path + '.params', 'rb'))

    # If Ain is specified
    if input_params['do_cnmfe']:
        Ain = None
        item_uuid = input_params['cnmfe_kwargs'].pop('Ain')

        if item_uuid:
            print('>> Ain specified, looking for cnm-A file <<')
            parent_batch_dir = os.environ['CURR_BATCH_DIR']
            item_out_file = os.path.join(parent_batch_dir, f'{item_uuid}.out')
            t0 = time()
            timeout = 60
            while not os.path.isfile(item_out_file):
                print('>>> cnm-A not found, waiting for 15 seconds <<<')
                sleep(15)
                if time() - t0 > timeout:
                    output.update({'status': 0, 'output_info': 'Timeout exceeding in waiting for Ain input file'})
                    raise TimeoutError('Timeout exceeding in waiting for Ain input file')

            if os.path.isfile(item_out_file):
                if json.load(open(item_out_file, 'r'))['status']:
                    Ain_path = os.path.join(parent_batch_dir, item_uuid + '_results.hdf5')
                    Ain = load_dict_from_hdf5(Ain_path)['estimates']['A']
                    print('>>> Found Ain file <<<')
                else:
                    raise FileNotFoundError('>>> Could not find specified Ain file <<<')

    print('*********** Creating Process Pool ***********')

    c, dview, n_processes = cm.cluster.setup_cluster(
        backend='local',
        n_processes=n_processes,
        single_thread=False,
        ignore_preexisting=True
    )

    try:
        print('Creating memmap')

        memmap_path = cm.save_memmap(
            filename,
            base_name=f'memmap-{UUID}',
            order='C',
            dview=dview,
            border_to_0=input_params['border_pix'],
        )

        Yr, dims, T = cm.load_memmap(memmap_path)
        Y = Yr.T.reshape((T,) + dims, order='F')

        if input_params['do_cnmfe']:
            gSig = input_params['cnmfe_kwargs']['gSig'][0]
        else:
            gSig = input_params['corr_pnr_kwargs']['gSig']

        cn_filter, pnr = cm.summary_images.correlation_pnr(
            Y, swap_dim=False, gSig=gSig
        )

        if not input_params['do_cnmfe'] and input_params['do_corr_pnr']:
            pickle.dump(cn_filter, open(UUID + '_cn_filter.pikl', 'wb'), protocol=4)
            pickle.dump(pnr, open(UUID + '_pnr.pikl', 'wb'), protocol=4)

            output_file_list = \
                [
                    UUID + '_pnr.pikl',
                    UUID + '_cn_filter.pikl',
                ]

            output.update(
                {
                    'output': UUID,
                    'status': 1,
                    'output_info': 'inspect correlation & pnr',
                    'output_files': output_file_list
                }
            )

            dview.terminate()

            for mf in glob(batch_dir + '/memmap-*'):
                try:
                    os.remove(mf)
                except:  # Windows doesn't like removing the memmaps
                    pass

            end_time = time()
            processing_time = (end_time - start_time) / 60
            output.update({'processing_time': processing_time})

            json.dump(output, open(file_path + '.out', 'w'))

            return

        cnmf_params_dict = \
            {
                "method_init": 'corr_pnr',
                "n_processes": n_processes,
                "only_init_patch": True,  # for 1p
                "center_psf": True,  # for 1p
                "normalize_init": False,  # for 1p
            }
        cnmf_params_dict.update(**input_params['cnmfe_kwargs'])

        cnm = cnmf.CNMF(
            n_processes=n_processes,
            dview=dview,
            Ain=Ain,
            params=CNMFParams(params_dict=cnmf_params_dict),
        )

        cnm.fit(Y)

        #  DISCARD LOW QUALITY COMPONENTS
        cnm.params.set(
            'quality',
            {
                'use_cnn': False,
                **input_params['eval_kwargs']
            }
        )

        cnm.estimates.evaluate_components(Y, cnm.params, dview=dview)

        out_filename = f'{UUID}_results.hdf5'
        cnm.save(out_filename)

        pickle.dump(pnr, open(UUID + '_pnr.pikl', 'wb'), protocol=4)
        pickle.dump(cn_filter, open(UUID + '_cn_filter.pikl', 'wb'), protocol=4)

        output.update(
            {
                'output': filename[:-5],
                'status': 1,
                'output_files': [
                    out_filename,
                    UUID + '_pnr.pikl',
                    UUID + '_cn_filter.pikl',
                ]

            }
        )

    except Exception as e:
        output.update({'status': 0, 'Y.shape': Y.shape, 'output_info': traceback.format_exc()})

    dview.terminate()

    if not input_params.get('keep_memmap', False):
        for mf in glob(batch_dir + '/memmap-*'):
            try:
                os.remove(mf)
            except: # Windows doesn't like removing the memmaps
                pass

    end_time = time()
    processing_time = (end_time - start_time) / 60
    output.update({'processing_time': processing_time})

    json.dump(output, open(file_path + '.out', 'w'))


class Output(QtWidgets.QWidget):
    def __init__(self, batch_dir, UUID, viewer_ref):
        # super(Output, self).__init__(self)
        # QtWidgets.QWidget.__init__(self, parent=None)
        self.batch_dir = batch_dir
        self.UUID = UUID
        self.viewer_ref = viewer_ref
        self.cnmfe_results = {}

        filename = os.path.join(batch_dir, str(UUID))

        if pickle.load(open(filename + '.params', 'rb'))['do_corr_pnr']:
            print('showing corr pnr')
            self.output_corr_pnr()
        else:
            QtWidgets.QWidget.__init__(self)
            print('opening question box')
            self.setWindowTitle('Show Correlation & PNR image or CNMFE')
            layout = QtWidgets.QVBoxLayout()

            label = QtWidgets.QLabel(self)
            label.setText('Would you like to look at the correlation & PNR image or view the CNMFE output?')
            layout.addWidget(label)

            self.btnCP = QtWidgets.QPushButton()
            self.btnCP.setText('Corr PNR Img')
            layout.addWidget(self.btnCP)

            self.btnImportIntoViewer = QtWidgets.QPushButton()
            self.btnImportIntoViewer.setText('Import CNMFE output into chosen Viewer')
            layout.addWidget(self.btnImportIntoViewer)

            self.checkbox_calc_raw_min_max = QtWidgets.QCheckBox()
            self.checkbox_calc_raw_min_max.setText('Calculate raw min and max of components')
            self.checkbox_calc_raw_min_max.setChecked(True)
            layout.addWidget(self.checkbox_calc_raw_min_max)

            self.setLayout(layout)

            self.btnCP.clicked.connect(self.output_corr_pnr)
            self.btnImportIntoViewer.clicked.connect(self.import_cnmfe_into_viewer)
            self.show()

    def output_corr_pnr(self):  # , batch_dir, UUID, viewer_ref):
        filename = os.path.join(self.batch_dir, str(self.UUID))

        cn_filter = pickle.load(open(filename + '_cn_filter.pikl', 'rb'))
        # try:
        pnr = pickle.load(open(filename + '_pnr.pikl', 'rb'))
        # except:
        #     try:
        #         pnr = pickle.load(open(filename + 'pnr.pikl', 'rb'))
            # except:
            #     return

        correlation_image_pnr = cn_filter
        pnr_image = pnr
        self.mw = MatplotlibWidget()
        # fig = mw.getFigure()
        self.mw.setWindowTitle(str(self.UUID))
        #Adapted from CaImAn library's visualization module
        # mw.plot = plt.figure(figsize=(10,4))
        # fig = pl.figure(figsize=(10, 4))
        self.cn_ax = self.mw.fig.add_axes([0.05, 0.2, 0.4, 0.7])
        self.cn_img = self.cn_ax.imshow(cn_filter, cmap='jet')
        # im_cn = plt.imshow(correlation_image_pnr, cmap='jet')
        # plt.title('correlation image')
        # self.cn_ax.colorbar()
        self.pn_ax = self.mw.fig.add_axes([0.5, 0.2, 0.4, 0.7])
        self.pn_img = self.pn_ax.imshow(pnr, cmap='jet')
        # im_pnr = plt.imshow(pnr_image, cmap='jet')
        # plt.title('PNR')
        # plt.colorbar()
        #
        s_cn_max_ax = self.mw.fig.add_axes([0.05, 0.01, 0.35, 0.03])
        s_cn_max = Slider(s_cn_max_ax, 'vmax',
                          correlation_image_pnr.min(), correlation_image_pnr.max(), valinit=correlation_image_pnr.max())

        s_cn_min_ax = self.mw.fig.add_axes([0.05, 0.07, 0.35, 0.03])
        s_cn_min = Slider(s_cn_min_ax, 'vmin',
                          correlation_image_pnr.min(), correlation_image_pnr.max(), valinit=correlation_image_pnr.min())

        s_pnr_max_ax = self.mw.fig.add_axes([0.5, 0.01, 0.35, 0.03])
        s_pnr_max = Slider(s_pnr_max_ax, 'vmax',
                           pnr_image.min(), pnr_image.max(), valinit=pnr_image.max())
        s_pnr_min_ax = self.mw.fig.add_axes([0.5, 0.07, 0.35, 0.03])
        s_pnr_min = Slider(s_pnr_min_ax, 'vmin',
                           pnr_image.min(), pnr_image.max(), valinit=pnr_image.min())
        #
        def update(val):
            self.cn_img.set_clim([s_cn_min.val, s_cn_max.val])
            self.pn_img.set_clim([s_pnr_min.val, s_pnr_max.val])
            self.mw.canvas.draw_idle()
        #
        s_cn_max.on_changed(update)
        s_cn_min.on_changed(update)
        s_pnr_max.on_changed(update)
        s_pnr_min.on_changed(update)
        # plt.ion()
        self.mw.show()
        # plt.show()

    def import_cnmfe_into_viewer(self):
        vi = ViewerUtils(self.viewer_ref)

        if not vi.discard_workEnv():
            return

        cnmf_data = load_dict_from_hdf5(
            os.path.join(self.batch_dir, f'{self.UUID}_results.hdf5')
        )

        vi.viewer.status_bar_label.showMessage('Loading CNMFE data, please wait...')
        pickle_file_path = os.path.join(self.batch_dir, f'{self.UUID}_input.pik')
        tiff_path = os.path.join(self.batch_dir, f'{self.UUID}_input.tiff')

        vi.viewer.workEnv = ViewerWorkEnv.from_pickle(pickle_file_path, tiff_path)
        vi.update_workEnv()

        input_params = pickle.load(
            open(
                os.path.join(self.batch_dir, f'{self.UUID}.params'),
                'rb'
            )
        )

        vi.viewer.workEnv.history_trace.append({'cnmfe': input_params})

        roi_manager_gui = vi.viewer.parent().get_module('roi_manager')
        roi_manager_gui.start_backend('CNMFROI')

        roi_manager_gui.manager.add_all_components(
            cnmf_data,
            input_params_dict=input_params,
            calc_raw_min_max=self.checkbox_calc_raw_min_max.isChecked()
        )

        vi.viewer.ui.label_curr_img_seq_name.setText('CNMFE of: ' + input_params['item_name'])
        self.close()


#if sys.argv[0] == __file__:
if __name__ == '__main__':
    run(sys.argv[1], sys.argv[2])



"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

import pandas as pd
import numpy as np
from typing import Tuple, List


def get_binary_stims(
        stim_df: pd.DataFrame,
        index_size: int,
        start_offset: int = 0,
        end_offset: int = 0
) -> Tuple[np.ndarray, List[str]]:
    """

    """

    stims = stim_df['name'].unique()

    if index_size < stim_df['end'].max():
        index_size = int(stim_df['end'].max())

    binarized_arrays = []

    for stimulus in stims:
        sub_df = stim_df[stim_df['name'] == stimulus]

        # One binary stim array for each stimulus in the `stim_df`
        stim_array = np.empty(index_size, dtype=np.uint8)
        stim_array[:] = 0

        # fill stim_array
        # stim_array is used for fancy indexing
        for ix, r in sub_df.iterrows():  # iterate through the stimulus periods
            # apply any offsets
            start_ix = r['start'] + start_offset
            end_ix = r['end'] + end_offset

            # set this period to 1 in the array
            stim_array[int(start_ix):int(end_ix)] = 1

        binarized_arrays.append(stim_array)

    return np.vstack(binarized_arrays), stims.tolist()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 5 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

import sys
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
print('Loading, please wait... ')
from mesmerize import configuration
from PyQt5.QtWidgets import QApplication
from mesmerize.common.window_manager import WindowManager
from mesmerize.project_manager import ProjectManager
from mesmerize.common.welcome_window import MainWindow
from mesmerize.scripts import *
from mesmerize import Transmission
from mesmerize.plotting import open_plot_file
import os
import logging
from datetime import datetime
from glob import glob
import click
import traceback
from zipfile import ZipFile, ZIP_LZMA


session_id = datetime.now().strftime("%Y-%m-%d_%H.%M.%S-%f")


class App(QApplication):
    window_manager = WindowManager()
    project_manager = ProjectManager()
    session_id = session_id


def start_welcome_window():
    app = App([])
    app.window_manager.welcome_window = MainWindow()
    app.window_manager.welcome_window.show()

    return app


def start_batch_manager(batch_path: str, item_uuid: str):
    app = App([])
    app.window_manager.welcome_window = MainWindow()
    app.window_manager.welcome_window.show()
    bm = app.window_manager.get_batch_manager(run_batch=[batch_path, item_uuid])
    return (app, bm)


def move_old_logfiles(logger, log_file_dir):
    zlogfile = ZipFile(
        os.path.join(log_file_dir, 'logs.zip'),
        mode='a',
        compression=ZIP_LZMA,
        allowZip64=True,

    )

    logfiles = glob(os.path.join(log_file_dir, '*.log'))
    logfiles.sort()
    while len(logfiles) > 5:
        logger.info("Moving old logfiles...")
        old_logfile = logfiles.pop(0)
        zlogfile.write(old_logfile, os.path.basename(old_logfile))
        os.remove(old_logfile)
    zlogfile.close()


@click.command()
@click.option('--log-level', type=str, default='INFO')
@click.option('--log-file', type=click.Path(exists=False, writable=True))
@click.option('--log-file-dir', type=click.Path(writable=True))
@click.option('--log-format', type=str)
@click.option('--run-batch', nargs=2, type=str)
@click.option('--open-plot', type=click.Path(exists=True, readable=True))
def main(
        log_level,
        log_file,
        log_file_dir,
        log_format,
        run_batch,
        open_plot
):
    if log_file_dir is None:
        log_file_dir = configuration.sys_cfg_dir

    if not os.path.exists(log_file_dir):
        os.makedirs(log_file_dir, exist_ok=True)

    if log_file is None:
        log_file = os.path.join(log_file_dir, f"{session_id}.log")

    if log_format is None:
        log_format = "%(asctime)s %(levelname)s %(pathname)s %(lineno)s \n %(message)s "

    logging.basicConfig(
        level=getattr(logging, log_level),
        format=log_format,
        filename=log_file,
        filemode='w'
    )

    def exception_hook(exc_type, exc_value, traceback):
        logging.error('EXCEPTION ENCOUNTERED', exc_info=(exc_type, exc_value, traceback))

    sys.excepthook = exception_hook

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    console.setFormatter(log_format)

    root_logger = logging.getLogger()

    root_logger.addHandler(console)
    root_logger.addHandler(
        logging.StreamHandler(sys.stdout)
    )
    root_logger.addHandler(
        logging.StreamHandler(sys.stderr)
    )

    if not configuration.IS_WINDOWS:
        move_old_logfiles(root_logger, log_file_dir)

    if not len(sys.argv) > 1:
        app = start_welcome_window()
        app.exec_()

    if run_batch:
        batch_path = run_batch[0]
        item = run_batch[1]
        app, bm = start_batch_manager(batch_path, item)
        app.exec_()

    elif open_plot:
        proj_path = os.path.dirname(
            os.path.dirname(open_plot)
        )
        app = start_welcome_window()

        app.window_manager.welcome_window.hide()
        app.project_manager.set(project_root_dir=proj_path)
        plot = open_plot_file(open_plot)
        plot.show()
        app.exec_()

    else:
        app = start_welcome_window()
        app.exec_()
    #
    # elif sys.argv[1] == 'lighten':
    #     create_lite_project.main(*sys.argv[2:])
    #
    # elif sys.argv[1] == 'show-graph':
    #     path = sys.argv[2]
    #     if not os.path.isfile(path):
    #         raise FileNotFoundError("File does not exist")
    #
    #     t = Transmission.from_hdf5(path)
    #     dbs = t.history_trace.data_blocks
    #     for db in dbs:
    #         t.history_trace.draw_graph(data_block_id=db, view=True)
    #


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtCore, QtWidgets
from ...variants import Heatmap
from ..base import BasePlotWidget
from .. import DatapointTracerWidget
from ....analysis.data_types import HistoryTrace, Transmission
from ....analysis import organize_dataframe_columns
from .control_widget_pytemplate import Ui_ControlWidget
import numpy as np
import pandas as pd
from typing import Optional, Union
from ....common.configuration import console_history_path
from ....common.qdialogs import present_exceptions, exceptions_label
from ....pyqtgraphCore.console import ConsoleWidget
from warnings import warn
import os


def help_func(e, tb):
    def show_info(info):
        QtWidgets.QMessageBox.information(None, 'Help info', info)
        return

    if str(e).startswith('all the input array dimensions for the concatenation axis must match exactly'):
        show_info('Make sure that the data under the "Data column" has been spliced.')

    elif str(e).startswith('Too many colors requested for the chosen cmap'):
        show_info('The chosen categorical labels column can only have a maximum of 20 unique labels.')

    elif 'unhashable type' in str(e):
        show_info('One of your selected columns has data type that is incompatible with what you have selected it for')


class HeatmapWidget(QtWidgets.QWidget):
    def __init__(self):
        QtWidgets.QWidget.__init__(self)

        self.vlayout = QtWidgets.QVBoxLayout(self)
        self.plot_widget = Heatmap()

        self.vlayout.addWidget(self.plot_widget)
        self.setLayout(self.vlayout)
        self.dataframe = pd.DataFrame()

    def set_data(self, array: np.ndarray = None, dataframe: pd.DataFrame = None, data_column: str = None,
                 cmap: str = 'jet'):

        if array is not None:
            assert isinstance(array, np.ndarray)
            data = array

        elif dataframe is not None:
            assert isinstance(dataframe, pd.DataFrame)
            assert isinstance(data_column, str)
            self.dataframe = dataframe
            data = np.vstack(self.dataframe[data_column].values)

        else:
            raise TypeError('You must pass either a numpy array or dataframe as and argument')

        self.plot_widget.set(data, cmap=cmap)


class HeatmapSplitterWidget(QtWidgets.QWidget):
    """Widget for interactive heatmaps"""
    def __init__(self, highlight_mode='row'):
        """

        :param highlight_mode: Interactive mode, one of 'row' or 'item'
        """
        QtWidgets.QWidget.__init__(self)
        self.vlayout = QtWidgets.QVBoxLayout(self)

        self.plot_variant: Heatmap = Heatmap(highlight_mode=highlight_mode)  #: That plot variant used by this plot widget

        self.labelSort = QtWidgets.QLabel(self)
        self.labelSort.setText('Sort heatmap according to column:')
        self.labelSort.setMinimumHeight(30)
        self.labelSort.setMaximumHeight(30)
        self.comboBoxSortColumn = QtWidgets.QComboBox(self)
        self.comboBoxSortColumn.currentTextChanged.connect(self._set_sort_order)

        self.plot_variant.layout().addWidget(self.labelSort)
        self.plot_variant.layout().addWidget(self.comboBoxSortColumn)

        self.splitter = QtWidgets.QSplitter(self)
        self.splitter.setStretchFactor(1, 1)
        self.splitter.addWidget(self.plot_variant)

        self.vlayout.addWidget(self.splitter)
        # self.vlayout.addSpacerItem(QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding))
        self.setLayout(self.vlayout)

        # self.plot_widget.signal_row_selection_changed.connect(self.set_current_datapoint)

        self.dataframe = None
        self.labels_column = None
        self.cmap = None

        self.previous_sort_column = ''

        self._transmission = None
        self._history_trace = None
        self.has_history_trace = False
        self.data_column = None

    def add_to_splitter(self, ui):
        self.splitter.addWidget(ui)
        return ui

    def _disconnect_comboBoxSort(self):
        try:
            self.comboBoxSortColumn.currentTextChanged.disconnect(self._set_sort_order)
        except TypeError:
            pass

    def _connect_comboBoxSort(self):
        self.comboBoxSortColumn.currentTextChanged.connect(self._set_sort_order)

    @staticmethod
    def merge_dataframes(dataframes) -> pd.DataFrame:
        if type(dataframes) is list:
            dataframe = pd.concat(dataframes)
        else:
            dataframe = dataframes

        if not isinstance(dataframe, pd.DataFrame):
            raise TypeError('Must pass a dataframe or list of dataframes')
        return dataframe

    def set_data(self, dataframes: Union[pd.DataFrame, list],
                 data_column: str,
                 labels_column: str,
                 cmap: str = 'jet',
                 transmission: Optional[Transmission] = None,
                 sort: bool = True,
                 reset_sorting: bool = True,
                 **kwargs):
        """
        Set the data and then set the plot

        :param dataframes:      list of dataframes or a single DataFrame
        :param data_column:     data column of the dataframe that is plotted in the heatmap
        :param labels_column:   dataframe column (usually categorical labels) used to generate the y-labels and legend.
        :param cmap:            colormap choice
        :param transmission:    transmission object that dataframe originates, used to calculate data units if passed
        :param sort:            if False, the sort comboBox is ignored
        :param reset_sorting:   reset the order of the rows in the heatmap
        :param kwargs:          Passed to Heatmap.set
        """
        dataframe = self.merge_dataframes(dataframes)
        self.data_column = data_column
        self.cmap = cmap
        self.labels_column = labels_column

        self.dataframe = dataframe.reset_index(drop=True)

        self.kwargs = kwargs
        # print(kwargs)

        self._transmission = None
        self._history_trace = None
        self.has_history_trace = False

        if transmission is not None:
            if not isinstance(transmission, Transmission):
                raise TypeError("Argument 'transmission' must be of type mesmerize.analysis.data_types.Transmission")
            self.set_transmission(transmission)

        self._disconnect_comboBoxSort()

        if reset_sorting:
            self.comboBoxSortColumn.clear()
            self.comboBoxSortColumn.addItems(organize_dataframe_columns(self.dataframe.columns)[1])

        if sort:
            ix = self.comboBoxSortColumn.findText(self.previous_sort_column)
            if (ix != -1) and (self.previous_sort_column != ''):
                self.comboBoxSortColumn.setCurrentIndex(ix)
                self.dataframe.sort_values(by=[self.previous_sort_column], inplace=True)

        data = np.vstack(self.dataframe[self.data_column].values)

        self._set_plot(data)

        self._connect_comboBoxSort()

    def _set_plot(self, data_array: np.ndarray):
        ylabels = self.dataframe[self.labels_column]
        self.plot_variant.set(data_array, cmap=self.cmap, ylabels=ylabels, **self.kwargs)

    @present_exceptions('Cannot sort', 'Make sure you choose an appropriate categorical column for sorting')
    def _set_sort_order(self, column: str):
        """
        Set the sort order of the heatmap rows according to a dataframe column. The column must contain categorical
        values. The rows are grouped together according to the categorical values.

        :param column: DataFrame column containing categorical values used for sorting the heatmap rows
        """
        self.dataframe.sort_values(by=[column], inplace=True)
        a = np.vstack(self.dataframe[self.data_column].values)
        self.previous_sort_column = column
        self._set_plot(a)

    def set_transmission(self, transmission: Transmission):
        """Set the input transmission"""
        self._transmission = transmission
        self.set_history_trace(transmission.history_trace)

    def get_transmission(self) -> Transmission:
        """Get the input transmission"""
        if self._transmission is None:
            raise ValueError('No tranmission has been set')
        else:
            return self._transmission

    def highlight_row(self, ix: int):
        """Highlight a row on the heatmap"""
        self.plot_variant.highlight_row(ix)

    def set_history_trace(self, history_trace: HistoryTrace):
        """Set the history trace object from the input transmission"""
        assert isinstance(history_trace, HistoryTrace)
        self._history_trace = history_trace
        self.has_history_trace = True


class ControlWidget(QtWidgets.QWidget):
    sig_changed = QtCore.pyqtSignal()

    def __init__(self):
        QtWidgets.QWidget.__init__(self)
        self.ui = Ui_ControlWidget()
        self.ui.setupUi(self)
        self.setMaximumWidth(350)

        self.ui.listWidgetColorMapsLabels.set_cmap('tab10')

        self.ui.listWidgetColorMapsData.signal_colormap_changed.connect(lambda: self.sig_changed.emit())
        self.ui.listWidgetColorMapsLabels.signal_colormap_changed.connect(lambda: self.sig_changed.emit())

        self.ui.comboBoxDataColumn.currentTextChanged.connect(lambda: self.sig_changed.emit())
        self.ui.comboBoxLabelsColumn.currentTextChanged.connect(lambda: self.sig_changed.emit())
        self.ui.comboBoxDPTCurveColumn.currentTextChanged.connect(lambda: self.sig_changed.emit())

        self.ui.pushButton_set_vmin_vmax.clicked.connect(self.sig_changed.emit)


class HeatmapTracerWidget(BasePlotWidget, HeatmapSplitterWidget):
    """Heatmap with an embedded datapoint tracer"""
    drop_opts = ['dataframes', 'transmission']  #: keys of the plot_opts dict that are not JSON compatible and not required for restoring this plot

    def __init__(self):
        super().__init__()

        self.setWindowTitle('Heatmap Tracer Widget')
        self.control_widget = ControlWidget()
        self.add_to_splitter(self.control_widget)

        self.live_datapoint_tracer = DatapointTracerWidget()  #: The embedded `Datapoint Tracer <API_DatapointTracer>`
        self.add_to_splitter(self.live_datapoint_tracer)

        ns = {'this': self,
              'get_plot_area': lambda: self.plot_variant
              }

        txt = ["Namespaces",
               "self as 'this'",
               "Usefull callables, see docs for details:",
               "get_plot_area()"
               ]

        txt = "\n".join(txt)

        cmd_history_file = os.path.join(console_history_path, 'heatmap_tracer_widget')

        self.console_widget = ConsoleWidget(parent=self, namespace=ns, text=txt, historyFile=cmd_history_file)
        self.console_widget.setVisible(False)
        self.control_widget.ui.pushButtonShowHideConsole.toggled.connect(self.console_widget.setVisible)
        self.vlayout.addWidget(self.console_widget)

        self.status_label = QtWidgets.QLabel(self)
        self.status_label.setMaximumHeight(32)
        self.status_label.mousePressEvent = self.show_exception_info
        self.vlayout.addWidget(self.status_label)

        self.exception_holder = None  #: Used for holding exceptions that can be viewed by clicking on self.status_label

        self.plot_variant.sig_selection_changed.connect(self.set_current_datapoint)

        self.datapoint_tracer_curve_column = None

        self.control_widget.ui.listWidgetColorMapsData.set_cmap('jet')

        self.control_widget.ui.pushButtonPlot.clicked.connect(self.update_plot)

        self.control_widget.ui.checkBoxLiveUpdate.toggled.connect(self.set_update_live)

        self.control_widget.ui.pushButtonSave.clicked.connect(self.save_plot_dialog)
        self.control_widget.ui.pushButtonLoad.clicked.connect(self.open_plot_dialog)

        self.update_live = False
        self.block_signals_list = [self.control_widget]

        self.is_clustering = False

        self.control_widget.sig_changed.connect(self.update_plot)

    @BasePlotWidget.signal_blocker
    def set_update_live(self, b: bool):
        """Set whether the plot should update live with changes in the flowchart"""
        self.update_live = b
        self.control_widget.ui.checkBoxLiveUpdate.setChecked(b)
        if b:
            self.update_plot()

    @QtCore.pyqtSlot(tuple)
    def set_current_datapoint(self, ix: tuple):
        """
        Set the currently selected datapoint in the :ref:`Datapoint Tracer <DatapointTracer>`.

        :param ix: index, (x, y). x is always 0 for this widget since it only uses 'row' selection mode and not 'item'
        """
        try:
            if self.is_clustering:
                ix = self.plot_variant.plot.dendrogram_row.reordered_ind[ix[1]]
            else:
                ix = ix[1]
            identifier = self.dataframe.iloc[ix]['uuid_curve']
        except IndexError:
            warn('Datapoint index out of bounds. Probably clicked a plot point outside of the data')
            return
        r = self.dataframe[self.dataframe['uuid_curve'] == identifier]

        if self.has_history_trace:
            block_id = r._BLOCK_
            if isinstance(block_id, pd.Series):
                block_id = block_id.item()
            h = self._history_trace.get_data_block_history(block_id)
        else:
            h = None

        self.live_datapoint_tracer.set_widget(datapoint_uuid=identifier,
                                              data_column_curve=self.datapoint_tracer_curve_column,
                                              row=r,
                                              proj_path=self.get_transmission().get_proj_path(),
                                              history_trace=h)

    @BasePlotWidget.signal_blocker
    def set_input(self, transmission: Transmission):
        """Set the input Transmission and update the plot if update_live is True"""
        if (self._transmission is None) or self.update_live:
            super(HeatmapTracerWidget, self).set_input(transmission)
            if self.update_live:
                self._set_vmin_vmax_limits()
                self.update_plot()

    def _set_vmin_vmax_limits(self):
        data_column = self.control_widget.ui.comboBoxDataColumn.currentText()
        data = np.vstack(self.transmission.df[data_column].values)

        _min = np.nanmin(data)
        _max = np.nanmax(data)

        range = abs(_min - _max)

        ws = [
                self.control_widget.ui.doubleSpinBox_vmin,
                self.control_widget.ui.doubleSpinBox_vmax,
             ]

        for w in ws:
            w.setMinimum(_min)
            w.setMaximum(_max)

        for w in ws:
            w.setSingleStep(range / 10)

        self.control_widget.ui.doubleSpinBox_vmin.setValue(_min)
        self.control_widget.ui.doubleSpinBox_vmax.setValue(_max)

    def fill_control_widget(self, data_columns: list, categorical_columns: list, uuid_columns: list):
        # self.control_widget.ui.comboBoxDataColumn.clear()
        self.control_widget.ui.comboBoxDataColumn.setItems(data_columns)

        # self.control_widget.ui.comboBoxLabelsColumn.clear()
        self.control_widget.ui.comboBoxLabelsColumn.setItems(categorical_columns)

        # self.control_widget.ui.comboBoxDPTCurveColumn.clear()
        self.control_widget.ui.comboBoxDPTCurveColumn.setItems(data_columns)

    def get_plot_opts(self, drop: bool = False) -> dict:
        """
        Get the plot options

        :param drop: Drop the non-json compatible objects that are not necessary to restore this plot
        """
        new_datacolumn = self.control_widget.ui.comboBoxDataColumn.currentText()

        if self.data_column != new_datacolumn:
            self._set_vmin_vmax_limits()

        d = dict(dataframes=self.transmission.df,
                 data_column=new_datacolumn,
                 labels_column=self.control_widget.ui.comboBoxLabelsColumn.currentText(),
                 datapoint_tracer_curve_column=self.control_widget.ui.comboBoxDPTCurveColumn.currentText(),
                 cmap=self.control_widget.ui.listWidgetColorMapsData.current_cmap,
                 transmission=self.transmission,
                 ylabels_cmap=self.control_widget.ui.listWidgetColorMapsLabels.current_cmap,
                 vmin=self.control_widget.ui.doubleSpinBox_vmin.value(),
                 vmax=self.control_widget.ui.doubleSpinBox_vmax.value()
                 )

        if drop:
            for k in self.drop_opts:
                d.pop(k)
        return d

    @BasePlotWidget.signal_blocker
    def set_plot_opts(self, opts: dict):
        """Set all plot options from a dict"""
        ix = self.control_widget.ui.comboBoxDataColumn.findText(opts['data_column'])
        self.control_widget.ui.comboBoxDataColumn.setCurrentIndex(ix)

        ix = self.control_widget.ui.comboBoxLabelsColumn.findText(opts['labels_column'])
        self.control_widget.ui.comboBoxLabelsColumn.setCurrentIndex(ix)

        ix = self.control_widget.ui.comboBoxDataColumn.findText(opts['datapoint_tracer_curve_column'])
        self.control_widget.ui.comboBoxDPTCurveColumn.setCurrentIndex(ix)

        self.control_widget.ui.listWidgetColorMapsData.set_cmap(opts['cmap'])
        self.control_widget.ui.listWidgetColorMapsLabels.set_cmap(opts['ylabels_cmap'])

    def update_plot(self):
        """Calls set_data and passes dict from get_plot_opts() as keyword arguments"""
        self.set_data(**self.get_plot_opts())

    def get_cluster_kwargs(self) -> dict:
        """Organize kwargs for visualization of hierarchical clustering"""
        # Just get the first datablock ID since Agglomerative clustering would have been done on all data blocks
        db_id = self.transmission.history_trace.data_blocks[0]
        linkage = np.array(self.transmission.history_trace.get_operation_params(data_block_id=db_id, operation='fcluster')['linkage_matrix'])
        cluster_labels = self.transmission.df['FCLUSTER_LABELS']
        ck = dict(row_linkage=linkage, row_cluster=True, col_cluster=False, cluster_labels=cluster_labels)
        return ck

    @exceptions_label('status_label', 'exception_holder', 'Error while setting data', 'Make sure you have selected appropriate columns')
    def set_data(self, *args, datapoint_tracer_curve_column: str = None, **kwargs):
        """
        Set the plot data, parameters and draw the plot.
        If the input Transmission comes directly from the FCluster it will pass a dict from get_cluster_kwargs() to the
        cluster_kwargs argument. Else it will pass None to cluster_kwargs.

        :param args: arguments to pass to superclass set_data() method
        :param datapoint_tracer_curve_column: Data column containing curves to use in the datapoint tracer
        :param kwargs:  keyword arguments, passed to superclass set_data() method
        """
        self.exception_holder = None
        self.status_label.clear()

        if self.transmission.last_output == 'fcluster':
            self.comboBoxSortColumn.setDisabled(True)
            self.is_clustering = True
            super(HeatmapTracerWidget, self).set_data(*args, cluster_kwargs=self.get_cluster_kwargs(), sort=False, **kwargs)
        else:
            self.is_clustering = False
            super(HeatmapTracerWidget, self).set_data(*args, cluster_kwargs=None, **kwargs)

        self.datapoint_tracer_curve_column = datapoint_tracer_curve_column

    def show_exception_info(self, mouse_press_ev):
        if self.exception_holder is not None:
            QtWidgets.QMessageBox.warning(self, *self.exception_holder)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './ui_files/cnmf_3d_pytemplate.ui'
#
# Created by: PyQt5 UI code generator 5.12
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_CNMF3D_template(object):
    def setupUi(self, CNMF3D_template):
        CNMF3D_template.setObjectName("CNMF3D_template")
        CNMF3D_template.resize(599, 735)
        CNMF3D_template.setFloating(True)
        self.dockWidgetContents = QtWidgets.QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.dockWidgetContents)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.groupBox_seed_components = QtWidgets.QGroupBox(self.dockWidgetContents)
        self.groupBox_seed_components.setCheckable(True)
        self.groupBox_seed_components.setChecked(False)
        self.groupBox_seed_components.setObjectName("groupBox_seed_components")
        self.verticalLayout_9 = QtWidgets.QVBoxLayout(self.groupBox_seed_components)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.lineEdit_seed_components_path = QtWidgets.QLineEdit(self.groupBox_seed_components)
        self.lineEdit_seed_components_path.setObjectName("lineEdit_seed_components_path")
        self.verticalLayout_9.addWidget(self.lineEdit_seed_components_path)
        self.verticalLayout_4.addWidget(self.groupBox_seed_components)
        self.label_6 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_6.setObjectName("label_6")
        self.verticalLayout_4.addWidget(self.label_6)
        self.spinBox_p = QtWidgets.QSpinBox(self.dockWidgetContents)
        self.spinBox_p.setProperty("value", 2)
        self.spinBox_p.setObjectName("spinBox_p")
        self.verticalLayout_4.addWidget(self.spinBox_p)
        self.groupBox_use_patches = QtWidgets.QGroupBox(self.dockWidgetContents)
        self.groupBox_use_patches.setFlat(False)
        self.groupBox_use_patches.setCheckable(True)
        self.groupBox_use_patches.setObjectName("groupBox_use_patches")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox_use_patches)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(self.groupBox_use_patches)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.spinBox_rf = QtWidgets.QSpinBox(self.groupBox_use_patches)
        self.spinBox_rf.setMaximum(9999)
        self.spinBox_rf.setSingleStep(5)
        self.spinBox_rf.setProperty("value", 25)
        self.spinBox_rf.setObjectName("spinBox_rf")
        self.verticalLayout.addWidget(self.spinBox_rf)
        self.label_2 = QtWidgets.QLabel(self.groupBox_use_patches)
        self.label_2.setObjectName("label_2")
        self.verticalLayout.addWidget(self.label_2)
        self.spinBox_stride = QtWidgets.QSpinBox(self.groupBox_use_patches)
        self.spinBox_stride.setMaximum(9999)
        self.spinBox_stride.setSingleStep(5)
        self.spinBox_stride.setProperty("value", 15)
        self.spinBox_stride.setObjectName("spinBox_stride")
        self.verticalLayout.addWidget(self.spinBox_stride)
        self.verticalLayout_4.addWidget(self.groupBox_use_patches)
        self.label_3 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_4.addWidget(self.label_3)
        self.spinBox_k = QtWidgets.QSpinBox(self.dockWidgetContents)
        self.spinBox_k.setMaximum(9999)
        self.spinBox_k.setProperty("value", 5)
        self.spinBox_k.setObjectName("spinBox_k")
        self.verticalLayout_4.addWidget(self.spinBox_k)
        self.label_4 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_4.addWidget(self.label_4)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.spinBox_gSig_x = QtWidgets.QSpinBox(self.dockWidgetContents)
        self.spinBox_gSig_x.setMaximum(999)
        self.spinBox_gSig_x.setProperty("value", 3)
        self.spinBox_gSig_x.setObjectName("spinBox_gSig_x")
        self.horizontalLayout.addWidget(self.spinBox_gSig_x)
        self.spinBox_gSig_y = QtWidgets.QSpinBox(self.dockWidgetContents)
        self.spinBox_gSig_y.setMaximum(999)
        self.spinBox_gSig_y.setProperty("value", 3)
        self.spinBox_gSig_y.setObjectName("spinBox_gSig_y")
        self.horizontalLayout.addWidget(self.spinBox_gSig_y)
        self.spinBox_gSig_z = QtWidgets.QSpinBox(self.dockWidgetContents)
        self.spinBox_gSig_z.setMaximum(999)
        self.spinBox_gSig_z.setProperty("value", 3)
        self.spinBox_gSig_z.setObjectName("spinBox_gSig_z")
        self.horizontalLayout.addWidget(self.spinBox_gSig_z)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout_4.addLayout(self.horizontalLayout)
        self.label_5 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_5.setObjectName("label_5")
        self.verticalLayout_4.addWidget(self.label_5)
        self.doubleSpinBox_merge_threshold = QtWidgets.QDoubleSpinBox(self.dockWidgetContents)
        self.doubleSpinBox_merge_threshold.setMinimum(0.01)
        self.doubleSpinBox_merge_threshold.setMaximum(1.0)
        self.doubleSpinBox_merge_threshold.setSingleStep(0.01)
        self.doubleSpinBox_merge_threshold.setProperty("value", 0.8)
        self.doubleSpinBox_merge_threshold.setObjectName("doubleSpinBox_merge_threshold")
        self.verticalLayout_4.addWidget(self.doubleSpinBox_merge_threshold)
        self.groupBox_cnmf_kwargs = QtWidgets.QGroupBox(self.dockWidgetContents)
        self.groupBox_cnmf_kwargs.setStatusTip("")
        self.groupBox_cnmf_kwargs.setCheckable(True)
        self.groupBox_cnmf_kwargs.setChecked(False)
        self.groupBox_cnmf_kwargs.setObjectName("groupBox_cnmf_kwargs")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.groupBox_cnmf_kwargs)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.plainTextEdit_cnmf_kwargs = QtWidgets.QPlainTextEdit(self.groupBox_cnmf_kwargs)
        self.plainTextEdit_cnmf_kwargs.setObjectName("plainTextEdit_cnmf_kwargs")
        self.verticalLayout_3.addWidget(self.plainTextEdit_cnmf_kwargs)
        self.verticalLayout_4.addWidget(self.groupBox_cnmf_kwargs)
        self.horizontalLayout_3.addLayout(self.verticalLayout_4)
        self.verticalLayout_6 = QtWidgets.QVBoxLayout()
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.label_7 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_7.setObjectName("label_7")
        self.verticalLayout_6.addWidget(self.label_7)
        self.doubleSpinBox_rval = QtWidgets.QDoubleSpinBox(self.dockWidgetContents)
        self.doubleSpinBox_rval.setMaximum(1.0)
        self.doubleSpinBox_rval.setSingleStep(0.01)
        self.doubleSpinBox_rval.setProperty("value", 0.7)
        self.doubleSpinBox_rval.setObjectName("doubleSpinBox_rval")
        self.verticalLayout_6.addWidget(self.doubleSpinBox_rval)
        self.label_8 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_8.setObjectName("label_8")
        self.verticalLayout_6.addWidget(self.label_8)
        self.doubleSpinBox_minSNR = QtWidgets.QDoubleSpinBox(self.dockWidgetContents)
        self.doubleSpinBox_minSNR.setSingleStep(0.25)
        self.doubleSpinBox_minSNR.setProperty("value", 2.5)
        self.doubleSpinBox_minSNR.setObjectName("doubleSpinBox_minSNR")
        self.verticalLayout_6.addWidget(self.doubleSpinBox_minSNR)
        self.label_9 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_9.setObjectName("label_9")
        self.verticalLayout_6.addWidget(self.label_9)
        self.doubleSpinBox_decay_time = QtWidgets.QDoubleSpinBox(self.dockWidgetContents)
        self.doubleSpinBox_decay_time.setMinimum(0.25)
        self.doubleSpinBox_decay_time.setProperty("value", 2.0)
        self.doubleSpinBox_decay_time.setObjectName("doubleSpinBox_decay_time")
        self.verticalLayout_6.addWidget(self.doubleSpinBox_decay_time)
        self.groupBox_eval_kwargs = QtWidgets.QGroupBox(self.dockWidgetContents)
        self.groupBox_eval_kwargs.setCheckable(True)
        self.groupBox_eval_kwargs.setChecked(False)
        self.groupBox_eval_kwargs.setObjectName("groupBox_eval_kwargs")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.groupBox_eval_kwargs)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.plainTextEdit_eval_kwargs = QtWidgets.QPlainTextEdit(self.groupBox_eval_kwargs)
        self.plainTextEdit_eval_kwargs.setObjectName("plainTextEdit_eval_kwargs")
        self.verticalLayout_5.addWidget(self.plainTextEdit_eval_kwargs)
        self.verticalLayout_6.addWidget(self.groupBox_eval_kwargs)
        self.checkBox_refit = QtWidgets.QCheckBox(self.dockWidgetContents)
        self.checkBox_refit.setObjectName("checkBox_refit")
        self.verticalLayout_6.addWidget(self.checkBox_refit)
        self.checkBox_keep_memmap = QtWidgets.QCheckBox(self.dockWidgetContents)
        self.checkBox_keep_memmap.setObjectName("checkBox_keep_memmap")
        self.verticalLayout_6.addWidget(self.checkBox_keep_memmap)
        self.groupBox_use_previous_memmap = QtWidgets.QGroupBox(self.dockWidgetContents)
        self.groupBox_use_previous_memmap.setCheckable(True)
        self.groupBox_use_previous_memmap.setChecked(False)
        self.groupBox_use_previous_memmap.setObjectName("groupBox_use_previous_memmap")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox_use_previous_memmap)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.lineEdit_memmap_uuid = QtWidgets.QLineEdit(self.groupBox_use_previous_memmap)
        self.lineEdit_memmap_uuid.setObjectName("lineEdit_memmap_uuid")
        self.verticalLayout_2.addWidget(self.lineEdit_memmap_uuid)
        self.verticalLayout_6.addWidget(self.groupBox_use_previous_memmap)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.lineEdit_name = QtWidgets.QLineEdit(self.dockWidgetContents)
        self.lineEdit_name.setObjectName("lineEdit_name")
        self.horizontalLayout_2.addWidget(self.lineEdit_name)
        self.pushButton_add_to_batch = QtWidgets.QPushButton(self.dockWidgetContents)
        self.pushButton_add_to_batch.setObjectName("pushButton_add_to_batch")
        self.horizontalLayout_2.addWidget(self.pushButton_add_to_batch)
        self.verticalLayout_6.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3.addLayout(self.verticalLayout_6)
        self.verticalLayout_7.addLayout(self.horizontalLayout_3)
        CNMF3D_template.setWidget(self.dockWidgetContents)

        self.retranslateUi(CNMF3D_template)
        self.lineEdit_name.returnPressed.connect(self.pushButton_add_to_batch.click)
        QtCore.QMetaObject.connectSlotsByName(CNMF3D_template)
        CNMF3D_template.setTabOrder(self.groupBox_seed_components, self.lineEdit_seed_components_path)
        CNMF3D_template.setTabOrder(self.lineEdit_seed_components_path, self.spinBox_p)
        CNMF3D_template.setTabOrder(self.spinBox_p, self.groupBox_use_patches)
        CNMF3D_template.setTabOrder(self.groupBox_use_patches, self.spinBox_rf)
        CNMF3D_template.setTabOrder(self.spinBox_rf, self.spinBox_stride)
        CNMF3D_template.setTabOrder(self.spinBox_stride, self.spinBox_k)
        CNMF3D_template.setTabOrder(self.spinBox_k, self.spinBox_gSig_x)
        CNMF3D_template.setTabOrder(self.spinBox_gSig_x, self.spinBox_gSig_y)
        CNMF3D_template.setTabOrder(self.spinBox_gSig_y, self.spinBox_gSig_z)
        CNMF3D_template.setTabOrder(self.spinBox_gSig_z, self.doubleSpinBox_merge_threshold)
        CNMF3D_template.setTabOrder(self.doubleSpinBox_merge_threshold, self.groupBox_cnmf_kwargs)
        CNMF3D_template.setTabOrder(self.groupBox_cnmf_kwargs, self.plainTextEdit_cnmf_kwargs)
        CNMF3D_template.setTabOrder(self.plainTextEdit_cnmf_kwargs, self.doubleSpinBox_rval)
        CNMF3D_template.setTabOrder(self.doubleSpinBox_rval, self.doubleSpinBox_minSNR)
        CNMF3D_template.setTabOrder(self.doubleSpinBox_minSNR, self.doubleSpinBox_decay_time)
        CNMF3D_template.setTabOrder(self.doubleSpinBox_decay_time, self.groupBox_eval_kwargs)
        CNMF3D_template.setTabOrder(self.groupBox_eval_kwargs, self.plainTextEdit_eval_kwargs)
        CNMF3D_template.setTabOrder(self.plainTextEdit_eval_kwargs, self.checkBox_refit)
        CNMF3D_template.setTabOrder(self.checkBox_refit, self.checkBox_keep_memmap)
        CNMF3D_template.setTabOrder(self.checkBox_keep_memmap, self.groupBox_use_previous_memmap)
        CNMF3D_template.setTabOrder(self.groupBox_use_previous_memmap, self.lineEdit_memmap_uuid)
        CNMF3D_template.setTabOrder(self.lineEdit_memmap_uuid, self.lineEdit_name)
        CNMF3D_template.setTabOrder(self.lineEdit_name, self.pushButton_add_to_batch)

    def retranslateUi(self, CNMF3D_template):
        _translate = QtCore.QCoreApplication.translate
        CNMF3D_template.setWindowTitle(_translate("CNMF3D_template", "&CNMF 3D"))
        self.groupBox_seed_components.setToolTip(_translate("CNMF3D_template", "Useful if trying out different paramters for the same input image stack."))
        self.groupBox_seed_components.setTitle(_translate("CNMF3D_template", "Seed components from masks"))
        self.lineEdit_seed_components_path.setPlaceholderText(_translate("CNMF3D_template", "Enter path to .npy file"))
        self.label_6.setText(_translate("CNMF3D_template", "p"))
        self.groupBox_use_patches.setTitle(_translate("CNMF3D_template", "&Use patches"))
        self.label.setText(_translate("CNMF3D_template", "rf"))
        self.label_2.setText(_translate("CNMF3D_template", "stride"))
        self.label_3.setText(_translate("CNMF3D_template", "k"))
        self.label_4.setText(_translate("CNMF3D_template", "gSig, (x, y z)"))
        self.label_5.setText(_translate("CNMF3D_template", "merge threshold"))
        self.groupBox_cnmf_kwargs.setToolTip(_translate("CNMF3D_template", "You can enter additional kwargs to pass for CNMF instantiation.\n"
"Use single quotes for strings, do not use double quotes."))
        self.groupBox_cnmf_kwargs.setTitle(_translate("CNMF3D_template", "Use CNMF &kwargs"))
        self.plainTextEdit_cnmf_kwargs.setToolTip(_translate("CNMF3D_template", "You can enter additional kwargs to pass for CNMF instantiation.\n"
"Use single quotes for strings, do not use double quotes."))
        self.label_7.setText(_translate("CNMF3D_template", "rval threshold"))
        self.label_8.setText(_translate("CNMF3D_template", "min SNR"))
        self.label_9.setText(_translate("CNMF3D_template", "decay time"))
        self.groupBox_eval_kwargs.setToolTip(_translate("CNMF3D_template", "You can enter additional parameters to use for component evaluation.\n"
"Use single quotes for strings, do not use double quotes"))
        self.groupBox_eval_kwargs.setTitle(_translate("CNMF3D_template", "Use e&valuation params"))
        self.plainTextEdit_eval_kwargs.setToolTip(_translate("CNMF3D_template", "You can enter additional parameters to use for component evaluation.\n"
"Use single quotes for strings, do not use double quotes"))
        self.checkBox_refit.setText(_translate("CNMF3D_template", "refit"))
        self.checkBox_keep_memmap.setToolTip(_translate("CNMF3D_template", "Keep the memmap of this batch item.\n"
"If unchecked the memmap will be deleted \n"
"after this batch item is processed unless it \n"
"is required for a later batch item."))
        self.checkBox_keep_memmap.setText(_translate("CNMF3D_template", "Keep memmap of this batch item"))
        self.groupBox_use_previous_memmap.setToolTip(_translate("CNMF3D_template", "Useful if trying out different paramters for the same input image stack."))
        self.groupBox_use_previous_memmap.setTitle(_translate("CNMF3D_template", "Use &memmap from previous batch item"))
        self.lineEdit_memmap_uuid.setPlaceholderText(_translate("CNMF3D_template", "Enter UUID"))
        self.lineEdit_name.setToolTip(_translate("CNMF3D_template", "Enter a name for this batch item"))
        self.lineEdit_name.setPlaceholderText(_translate("CNMF3D_template", "Enter name"))
        self.pushButton_add_to_batch.setText(_translate("CNMF3D_template", "Add to batch"))





#!/usr/bin/env python3:
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 22 19:01:21 2017

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007

Just a clean simple & independent image data class which is the core of the work
envrionment in Mesmerize Viewer.

seq : 3D array (2D + time) of the image sequence
meta : meta data dictionary
Map : Stimulus map. Contains the definitions of the stimuli & the time that they
occured for the animal that was exposed to in this particular image sequence

"""
import numpy as np


class ImgData:
    """Object that stores the image sequence and meta data from the imaging source"""
    def __init__(self, seq: np.ndarray = None, meta: dict = None):
        """
        :param seq:     Image sequence as a numpy array, shape is [x, y, t] or [x, y, t, z]
        :param meta:    Meta data dict from the imaging source.
        """

        self.seq = None
        self._seq = None

        self.z = None
        self.z_max = None
        self._meta = None

        if seq is not None:
            self.ndim = seq.ndim

            if seq.ndim == 4:
                self._seq = seq
                self.seq = self._seq[:, :, :, 0]
                self.z = 0
                self.z_max = self._seq.shape[3] - 1
            else:
                self._seq = seq
                self.seq = self.seq = self._seq  #: image sequence, shape is [x, y, t] or [x, y, t, z]
        else:
            self.ndim = None

        if meta is None:
            meta = {'origin': 'unknown',
                    'fps': 0,
                    'date': 'unknown',
                    'orig_meta': None
                    }

        self.meta = meta.copy()  #: Meta data dict from the imaging source

    @property
    def meta(self) -> dict:
        return self._meta

    @meta.setter
    def meta(self, m: dict):
        reqs = ['origin', 'fps', 'date']
        if not all(k in m.keys() for k in reqs):
            raise ValueError(f"Meta data dict must contain all of the following mandatory fields:\n{reqs}")

        self._meta = m

    def set_zlevel(self, z):
        if self._seq.ndim < 4:
            raise ValueError('Data are not 3D, cannot set z-level')

        self.z = z
        self.seq = self._seq[:, :, :, self.z]

    def clear(self):
        del self.seq
        del self._seq
        del self._meta

        self.seq = None
        self._seq = None
        self.meta = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on March 29 2018

@author: kushal, daniel

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007

Just a bunch of generalized classes for image acquisition, preview, and writing from Hamamatsu and OpenCV cameras.

"""

import sys
from .hamamatsu import hamamatsu_camera
import abc
import threading
import Queue
import cv2
import numpy as np
import pyqtgraph as pg
import time
import datetime
import json
import tifffile
import multiprocessing
import traceback
import pickle


class BaseCamera(threading.Thread):
    """Base class for all types of cameras"""
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        threading.Thread.__init__(self)

    @abc.abstractmethod
    def run(self):
        pass

    @abc.abstractmethod
    def end(self):
        pass

    @property
    def framerate(self):
        return self._framerate

    @framerate.setter
    def framerate(self, framerate):
        self._framerate = framerate


class BasePreview:
    """Base class to display a preview of a frame from the camera using the ImageView class from pyqtgraph"""
    def __init__(self):
        self.iv = pg.imageview.ImageView()

        colors = [
            (0, 0, 0),
            (7, 0, 220),
            (236, 0, 134),
            (246, 246, 0),
            (255, 255, 255),
            (0, 255, 0)
        ]

        cmap = pg.ColorMap(pos=np.linspace(0.0, 1.0, 6), color=colors)
        self.iv.setColorMap(cmap)
        self.iv.setLevels(0, 65535)
        self.hist = self.iv.getHistogramWidget()
        self.hist.vb.enableAutoRange(self.hist.vb.YAxis, False)
        self.iv.show()

    def update_preview(self, img):
        self.iv.setImage(img, autoRange=False, autoLevels=False, autoHistogramRange=False)


class ImageFuncs:
    """Misc image functions"""
    @staticmethod
    def create_LUT_8bit(levels, src_depth):
        """
        :param src_depth:   bit depth of the source
        :type src_depth:    int
        :param levels:      min and max levels with which to create the LUT
        :type levels:       tuple
        :return:            8 bit LUT to convert the source's bit depth to 8 bit
        :rtype:             np.ndarray
        """

        accepted_depths = [16, 32, 64]
        if src_depth not in accepted_depths:
            raise TypeError('Can only convert from uint16, uint32, or uint64')
        depth = 'uint' + str(src_depth)
        LUT = np.arange(2**src_depth, dtype=depth)
        LUT.clip(levels[0], levels[1], out=LUT)
        LUT -= levels[0]
        print("LUT is: ")
        print(LUT)
        print("levels is: ")
        print(levels)
        print("denominator is")
        print((levels[1] - levels[0] + 1) / 256)
#        pickle.dump(LUT, open('F:\LUT.pickle', 'wb'))
#        pickle.dump(levels, open('F:\levels.pickle', 'wb'))
        np.floor_divide(LUT, (levels[1] - levels[0] + 1) / 256, out=LUT, casting='unsafe')
            

        return LUT.astype(np.uint8)

    @staticmethod
    def apply_8bit_LUT(image, LUT):
        """
        :param image:   The image upon which to apply the LUT and change its bit depth
        :type image:    np.ndarray
        :param LUT:     The 8bit LUT to use for downscaling the bit depth. Generated by ImageFuncs.create_LUT_8bit
        :type LUT:      np.ndarray
        :return:        Downscaled 8 bit image with the LUT applied to it
        :rtype:         np.ndarray
        """
        return np.take(LUT, image).astype(np.uint8)


class BaseWriter(threading.Thread):
    """Base class that contains methods for converting 16 bit to 8 bit grey-scale frames and saving metadata"""
    def __init__(self, threading_queue, filename, compression_level=1, levels=(0, 65535), metadata={}):
        threading.Thread.__init__(self)
        assert isinstance(threading_queue, Queue.Queue)
        self.q = threading_queue
        self.filename = filename
        self.comp_lev = compression_level
        self.levels = levels
        self.metadata = metadata
        if 'framerate' not in self.metadata.keys():
            try:
                self.metadata['framerate'] = 1/self.metadata['exposure']
            except KeyError:
                raise KeyError('Exposure or Framerate must be specified to save as metadata')

        self.tiff_writer = tifffile.TiffWriter(filename, bigtiff=True, append=True)

    @abc.abstractmethod
    def run(self):
        pass

    @abc.abstractmethod
    def end(self):
        pass

    # def create_LUT_8bit(self, src_depth):
    #     """
    #     :param src_depth:   bit depth of the source
    #     :return:            8 bit LUT to convert the source's bit depth to 8 bit
    #     :rtype:             np.ndarray
    #     """
    #     accepted_depths = [16, 32, 64]
    #     if src_depth not in accepted_depths:
    #         raise TypeError('Can only convert from uint16, uint32, or uint64')
    #     depth = 'uint' + str(src_depth)
    #     LUT = np.arange(2**src_depth, dtype=depth)
    #     LUT.clip(self.levels[0], self.levels[1], out=LUT)
    #     LUT -= self.levels[0]
    #     np.floor_divide(LUT, (self.levels[1] - self.levels[0] + 1) / 256,
    #                     out=LUT, casting='unsafe')
    #     return LUT.astype(np.uint8)
    #
    # def apply_8bit_LUT(self, image, LUT):
    #     return np.take(LUT, image).astype(np.uint8)

    def save_metadata(self, filename):
        date = datetime.datetime.fromtimestamp(time.time())
        ymd = date.strftime('%Y%m%d')
        hms = date.strftime('%H%M%S')

        meta = {'framerate':    self.metadata['framerate'],
                'source':       'AwesomeImager',
                'version':      self.metadata['version'],
                'date':         ymd,
                'time':         hms,
                'stims':        self.metadata['stims'],
                'level_min':    int(self.metadata['levels'][0]),
                'level_max':    int(self.metadata['levels'][1])}

        if filename.endswith('.tiff'):
            json_file = filename[:-5] + '.json'
        elif filename.endswith('.tif'):
            json_file = filename[:-4] + '.json'
        else:
            raise ValueError
        with open(json_file, 'w') as f:
            json.dump(meta, f)


class BaseHamamatsu(BaseCamera):
    """Base class for Hamamatsu cameras"""
    __metaclass__ = abc.ABCMeta

    def __init__(self, **kwargs):
        BaseCamera.__init__(self)
        if 'exposure' not in kwargs.keys():
            raise KeyError('No exposure value specified')

        self.hcam = hc.HamamatsuCameraMR(0)
        self.camera_open = True

        cam_offset = 100
        cam_x = 2048
        cam_y = 2048
        self.hcam.setPropertyValue("defect_correct_mode", "OFF")
        self.hcam.setPropertyValue("subarray_hsize", cam_x)
        self.hcam.setPropertyValue("subarray_vsize", cam_y)
        self.hcam.setPropertyValue("binning", "2x2")
        self.hcam.setPropertyValue("readout_speed", 2)
        self.exposure = kwargs['exposure']

    @property
    def exposure(self):
        e = self.hcam.getPropertyValue("exposure_time")
        return e

    @exposure.setter
    def exposure(self, e):
        self.hcam.setPropertyValue("exposure_time", e)
        self.framerate = 1/e

    def get_grey_values(self):
        """
        :rtype: np.ndarray
        :return: 1D numpy array of grey values
        """

        [frame, dim] = self.hcam.getFrames()
        grey_values = frame[0].getData()
        return grey_values

    @abc.abstractmethod
    def end(self):
        self.hcam.stopAcquisition()
        self.hcam.shutdown()
        self.camera_open = False


class BaseOpenCV(BaseCamera):
    """
    Adapted from Daniel Dondorp
    Base class for OpenCV compatible cameras
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, camera=0, framerate=30, shape=(7680, 4320), exposure=-5.0):
        """
        :param camera: Which connected camera to use.
        :param framerate:
        :param shape:
        :param brightness:
        :param exposure: negative values, get evaluates  as 2**param_val by camera.
        """
        BaseCamera.__init__(self)
        self.camera = camera
        self.cap = cv2.VideoCapture(self.camera)
        self.alive = False
        self.framerate = framerate
        self.shape = shape

        self.aperture = self.cap.get(cv2.CAP_PROP_APERTURE)
        self.exposure = exposure

    @property
    def framerate(self):
        return self._framerate

    @framerate.setter
    def framerate(self, framerate):
        self._framerate = framerate
        self.cap.set(cv2.CAP_PROP_FPS, framerate)

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, shape):
        self._shape = shape
        w, h = shape
        self.cap.set(cv2.CAP_PROP_FRAME_WIDTH, w)
        self.cap.set(cv2.CAP_PROP_FRAME_HEIGHT, h)
        # print("Video shape set to : ", w, h)

    @property
    def exposure(self):
        return self._exposure

    @exposure.setter
    def exposure(self, exposure):
        self._exposure = exposure
        self.cap.set(cv2.CAP_PROP_EXPOSURE, exposure)

    def reconnect(self):
        self.cap.release()
        self.cap = cv2.VideoCapture(self.camera)

    def read(self):
        ret = False
        while ret == False:
            ret, frame = self.cap.read()

        return cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

    def end(self):
        self.cap.release()


class PreviewOpenCV(BaseOpenCV, BasePreview):
    """
    Adapted from Daniel Dondorp
    """
    def __init__(self):
        BaseOpenCV.__init__(self)
        BasePreview.__init__(self)
        super(PreviewOpenCV, self).__init__(BasePreview)

    def run(self):
        print("Preview Starting")
        self.alive = True

        frames = 0
        start_time = time.time()

        while self._alive:

            ret, frame = self.cap.read()

            if ret:
                self.update_preview(frame)

            frames += 1
            if frames == 10:
                end_time = time.time()
                total_time = (end_time - start_time)
                fps = frames / total_time
                sys.stdout.write("\r fps = " + str(fps) + " time for 10 frames: " + str(total_time))
                frames = 0
                start_time = time.time()
        self.iv.close()
        super(PreviewOpenCV, self).end()

    def end(self):
        self._alive = False


class AcquireOpenCV(BaseOpenCV):
    """
    Adapted from Daniel Dondorp
    """
    def __init__(self):
        BaseOpenCV.__init__(self)

    def run(self, name="out", duration=60):
        w = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        h = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        framerate = self.framerate

        savepath = name + ".avi"

        fourcc = cv2.VideoWriter_fourcc(*"XVID")
        out = cv2.VideoWriter(savepath, fourcc, framerate, (w, h), isColor=False)

        start_time = time.time()
        end_time = start_time
        self.alive = True
        fc = 0
        while (end_time - start_time) < duration and self.alive:

            ret, frame = self.cap.read()

            if ret:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
                out.write(frame)
                fc += 1
                cv2.imshow('Recording', frame)

            end_time = time.time()
            sys.stdout.write("\r Recording! " + str(np.round((end_time - start_time), 2)) + "/" + str(
                duration) + " seconds       ")

        else:
            sys.stdout.write("\n Recording Complete! Saved as " + name + "Framecount: " + str(fc))

            cv2.destroyWindow("Recording")
            out.release()
            self.alive = False


class PreviewHamamatsu(BaseHamamatsu, BasePreview):
    def __init__(self, **kwargs):
        BaseHamamatsu.__init__(self, **kwargs)
        BasePreview.__init__(self)
        self._show_preview = True

    def run(self):
        self.hcam.startAcquisition()
        first_img = True
        while self._show_preview:
            try:
                img = np.reshape(self.get_grey_values(), (1024, 1024))
                self.iv.setImage(img, autoRange=False, autoLevels=False, autoHistogramRange=False)

                if first_img:
                    self.iv.autoLevels()
                    first_img = False
                self.levels = self.hist.getLevels()

            except Exception as e:
                print(e)

        super(PreviewHamamatsu, self).end()

    def end(self):
        self._show_preview = False
        self.iv.close()


class AcquireHamamatsu(BaseHamamatsu):
    def __init__(self, parameters, threading_queue, duration):
        kwargs = parameters
        BaseHamamatsu.__init__(self, **kwargs)
        self.q = threading_queue
        self.duration = duration
        self._acquire = True

    def run(self):
        frame_num = 0
        self.hcam.startAcquisition()
        self.stop_time = time.time() + self.duration
        while (time.time() < self.stop_time) and self._acquire:
            try:
                # self.lens.focalpower( <<appropriate focal power>> )

                # << Modulo calculation to find the next focal power to adjust to according to focal_interval >>

                # time.sleep(0.014) # Wait 14ms for the lens to adjust to the right focal power

                # <<** might be possible to lower this time because of small stack intervals
                # and use larger wait times only when beginning the next stack!! ** >>

                self.q.put(self.get_grey_values())
                print('Read frame num ' + str(frame_num) + '\r')
                frame_num += 1

            except KeyboardInterrupt:
                self.end()
                raise KeyboardInterrupt

            except Exception as e:
                if type(e) is not IndexError:
                    print('Something went wrong during ' +
                    'acquisition of frame num: ' + str(frame_num) 
                    + '\n' + str(traceback.format_exc()))

        self.q.put('done')
        super(AcquireHamamatsu, self).end()

    def end(self):
        self._acquire = False


# TODO: TRY MULTIPROCESSING HERE SINCE I'M ONLY SENDING A NUMPY ARRAY AND NOT HCAM OBJECT !!!!!!!!!!
class WriterHamamatsu(BaseWriter, BasePreview):
    """Use for Hamamatsu Cameras"""

    def __init__(self, threading_queue, filename, compression_level, levels, metadata):
        BaseWriter.__init__(self, threading_queue=threading_queue, filename=filename,
                            compression_level=compression_level, levels=levels, metadata=metadata)
        BasePreview.__init__(self)
        self.iv.setLevels(levels[0], levels[1])
        self.LUT_8bit = ImageFuncs.create_LUT_8bit(levels, 16)
        self.write = True

    def run(self):
        img_num = 0
        while self.write:
            if not self.q.not_empty:
                continue

            cam_data = self.q.get()

            if type(cam_data) is str:
                if cam_data == 'done':
                    break

            else:
                try:
                    img = np.reshape(cam_data, (1024, 1024))
                    try:
                        self.iv.setImage(img, autoRange=False, autoLevels=False, autoHistogramRange=False)
                    except Exception as e:
                        print('Error displaying an image: ' + str(e))

                    #img = ImageFuncs.apply_8bit_LUT(img, self.LUT_8bit)

                    self.tiff_writer.save(img, compress=self.comp_lev)
                    print('qsize is: ' + str(self.q.qsize()) + '\r')
                    print('wrote ImgNum: ' + str(img_num)  + '\r')

                    self.q.task_done()
                    img_num += 1

                except KeyboardInterrupt:
                    break

        self.tiff_writer.close()
        self.save_metadata(self.filename)

    def end(self):
        self.write = False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2019 Kushal Kolar
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from fcsugar import *
from ..containers import TransmissionContainer
import numpy as np
from ...utils import pad_arrays as _pad_arrays


@node
def pad_arrays(container, data_column, method: str = 'random'):
    data = _pad_arrays(container.df[data_column].values, method=method)

    container.df['pad_arrays'] = data.tolist()
    container.df['pad_arrays'] = container.df['pad_arrays'].apply(np.array)

    return container


@node
def partition(container, n_partitions: int):
    size = container.df.index.size

    ixs = np.array(range(size))

    partitions = np.array_split(ixs, n_partitions)
    labels = np.empty(shape=(size,), dtype=np.int64)

    for label, p in enumerate(partitions):
        labels[p] = label

    container.df['partition'] = labels
    return container

# TODO: Create container for numpy arrays!
@node
def sample_partition(contrainer, data_column: str, labels_column: str) -> np.ndarray:
    partitions = contrainer.df[labels_column].unique()

    df = contrainer.df

    samples = []
    for p in partitions:
        s = df[data_column][df[labels_column] == p].sample(1).values[0]
        samples.append(s)

    samples = np.array(samples)

    return samples



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Apr 27 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from psutil import cpu_count
from . import configuration
from .pytemplates.system_config_pytemplate import *
from functools import partial
import traceback
import os


class SystemConfigGUI(QtWidgets.QWidget):
    def __init__(self, parent=None, *args):
        QtWidgets.QWidget.__init__(self, parent, *args)
        self.setWindowTitle('System Configuration')
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.ui.spinBoxThreads.setMaximum(cpu_count() - 1)

        self.set_ui_from_config()

        self.ui.btnApply.clicked.connect(self.set_config_from_ui)
        self.ui.btnClose.clicked.connect(self.close)

        self.ui.pushButtonCUDAError.setVisible(False)
        self.ui.pushButtonCUDAError.setDisabled(True)

        self.check_cuda()

        self.ui.btnReloadConfigFile.clicked.connect(self.set_ui_from_config)
        self.ui.pushButtonResetDefaults.clicked.connect(
            partial(self.set_ui_from_config, configuration.default_sys_config))

        self.saved = True

    def _set_config(self, **kwargs):
        sys_cfg = configuration.get_sys_config()
        for k in kwargs.keys():
            sys_cfg['_MESMERIZE_{}'.format(k.upper())] = kwargs[k]
        configuration.save_sys_config(sys_cfg)

    def set_ui_from_config(self, config: dict = None):
        if config is None:
            config = configuration.get_sys_config()
        n_threads = config['_MESMERIZE_N_THREADS']
        python_call = config['_MESMERIZE_PYTHON_CALL']
        use_cuda = config['_MESMERIZE_USE_CUDA']
        workdir = config['_MESMERIZE_WORKDIR']
        prefix_commands = config['_MESMERIZE_PREFIX_COMMANDS']
        custom_modules_dir = config['_MESMERIZE_CUSTOM_MODULES_DIR']

        self.ui.spinBoxThreads.setValue(n_threads)
        self.ui.lineEditPythonCall.setText(python_call)
        self.ui.checkBoxUseCUDA.setChecked(use_cuda)
        self.ui.lineEditWorkDir.setText(workdir)
        self.ui.plainTextEditPrefixCommands.setPlainText(prefix_commands)
        self.ui.lineEditCustomModulesDir.setText(custom_modules_dir)

    def _get_ui_config(self) -> dict:
        return dict(n_threads=self.ui.spinBoxThreads.value(),
                    python_call=self.ui.lineEditPythonCall.text(),
                    use_cuda=self.ui.checkBoxUseCUDA.isChecked(),
                    workdir=self.ui.lineEditWorkDir.text(),
                    prefix_commands=self.ui.plainTextEditPrefixCommands.toPlainText(),
                    custom_modules_dir=self.ui.lineEditCustomModulesDir.text())

    def set_config_from_ui(self):
        if self.check_workdir():
            self._set_config(**self._get_ui_config())
        else:
            QtWidgets.QMessageBox.warning(self, 'Insufficient permissions',
                                          'You do not have write access to the chosen work directory.\n'
                                          'Enter a different directory')
            return

    def check_workdir(self) -> bool:
        workdir = self.ui.lineEditWorkDir.text()
        if os.access(workdir, os.W_OK):
            return True
        else:
            return False

    def check_cuda(self) -> bool:
        self.cuda_error_msg = None

        try:
            import pycuda
            import skcuda
            from pycuda import driver
            import pycuda.autoinit
            driver.Device.count()
            return True

        except ImportError:
            self.cuda_error_msg = traceback.format_exc()
            self.ui.checkBoxUseCUDA.setText('Use CUDA - Disabled, could not import libraries')

        except driver.LogicError:
            self.cuda_error_msg = traceback.format_exc()
            self.ui.checkBoxUseCUDA.setText('Use CUDA - Disabled, could not find CUDA devices using driver')

        self.disable_cuda_UI()
        return False

    def disable_cuda_UI(self):
        self.ui.checkBoxUseCUDA.setChecked(False)
        self.ui.checkBoxUseCUDA.setDisabled(True)
        self.ui.pushButtonCUDAError.setVisible(True)
        self.ui.pushButtonCUDAError.setEnabled(True)
        self.ui.pushButtonCUDAError.clicked.connect(self.show_cuda_error)

    def show_cuda_error(self):
        QtWidgets.QMessageBox.information(self, 'CUDA Error', self.cuda_error_msg)

    # def close_requested(self):
    #     cfg = self._get_ui_config()
    #     sys_cfg = configuration.get_sys_config()
    #     for k in cfg.keys():
    #         if sys_cfg[k] != cfg[k]:
    #             if QtWidgets.QMessageBox.warning(self, 'Unsaved changes',
    #                                          'You have not applied your changes, '
    #                                          'would you like to close anyways?',
    #                                          QtWidgets.QMessageBox.Yes,
    #                                          QtWidgets.QMessageBox.No) == QtWidgets.QMessageBox.No:
    #                 return
    #             else:
    #                 return

    # def show(self):
    #     self.load_ui_from_config_file()
    #     super(SystemConfigGUI, self).show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri March 2 2017

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

import numpy as np
from scipy.integrate import simps
from .data_types import Transmission
from ..common.configuration import IS_WINDOWS
from ..common import get_sys_config
from tqdm import tqdm
import pandas as pd
import uuid
from warnings import warn

if not IS_WINDOWS:
    from multiprocessing import Pool
else:
    from multiprocessing.pool import ThreadPool as Pool


feature_list = \
[
    '_pf_ampl_rel_b_ix_l',
    '_pf_ampl_rel_b_ix_r',
    '_pf_ampl_rel_b_mean',
    '_pf_ampl_rel_zero',
    '_pf_area_rel_zero',
    '_pf_area_rel_min',
    '_pf_rising_slope_avg',
    '_pf_falling_slope_avg',
    '_pf_duration_base',
    '_pf_peak_curve',
    '_pf_uuid',
    '_pf_p_ix',
    '_pf_b_ix_l',
    '_pf_b_ix_r'
]

nan_dict = {k: np.nan for k in feature_list}


class ComputePeakFeatures:
    def compute(self, transmission: Transmission, data_column: str):
        self.t = transmission
        self.t.df.reset_index(drop=True, inplace=True)
        self.data_column = data_column
        with Pool(get_sys_config()['_MESMERIZE_N_THREADS']) as pool:
            dfs = list(tqdm(pool.imap(self._per_curve, self.t.df.iterrows()), total=self.t.df.index.size))

        self.t.df = pd.concat([df for df in dfs if df is not None]).reset_index(drop=True)

        return self.t

    def _per_curve(self, row: pd.Series):
        row = row[1]
        pb_df = row['peaks_bases']

        curve = row[self.data_column]

        out_df = pb_df[pb_df.label == 'peak']

        if out_df.empty:
            return None

        pd.options.mode.chained_assignment = None

        out_df[
            [
                '_pf_ampl_rel_b_ix_l',
                '_pf_ampl_rel_b_ix_r',
                '_pf_ampl_rel_b_mean',
                '_pf_ampl_rel_zero',
                '_pf_area_rel_zero',
                '_pf_area_rel_min',
                '_pf_rising_slope_avg',
                '_pf_falling_slope_avg',
                '_pf_duration_base',
                '_pf_peak_curve',
                '_pf_uuid',
                '_pf_p_ix',
                '_pf_b_ix_l',
                '_pf_b_ix_r'
            ]

        ] \
            = out_df.apply(lambda r: self._per_peak(r, curve, pb_df), axis=1)

        row_dup = pd.DataFrame([row] * out_df.index.size)
        out_df = pd.concat([row_dup.reset_index(drop=True), out_df.reset_index(drop=True)], axis=1)
        return out_df

    def _per_peak(self, row, curve, pb_df):
        p_ix = row.event  #: Peak index relative to the whole curve
        ix = row.name

        if ix == (pb_df.index.size - 1) or ix == 0:
            warn(f"Peak at curve index: <{p_ix}> is not flanked by bases on both sides, ignoring")
            return pd.Series(nan_dict)

        try:
            if pb_df.iloc[ix - 1].label != 'base' and pb_df.iloc[ix + 1].label != 'base':
                warn(f"Peak at curve index: <{p_ix}> is not flanked by bases on both sides, ignoring")
                return pd.Series(nan_dict)

        except IndexError as e:
            warn(f"Index error with a peak, probably missing a flanking base:\n{e}\nignoring")
            return pd.Series(nan_dict)

        b_ix_l = pb_df.iloc[ix - 1].event  #: Left base index relative to the whole curve
        b_ix_r = pb_df.iloc[ix + 1].event  #: Right base index relative to the whole curve

        max_ix = len(curve) - 1

        if (b_ix_l > max_ix) or (b_ix_r > max_ix) or (p_ix > max_ix):
            warn(f'Base or peak index out of bounds for peak at curve index: <{p_ix}>, ignoring')
            return pd.Series(nan_dict)

        peak_curve = curve[b_ix_l:b_ix_r]

        out = {'_pf_ampl_rel_b_ix_l': curve[p_ix] - curve[b_ix_l],
               '_pf_ampl_rel_b_ix_r': curve[p_ix] - curve[b_ix_r],
               '_pf_ampl_rel_b_mean': curve[p_ix] - np.mean((curve[b_ix_l], curve[b_ix_r])),
               '_pf_ampl_rel_zero': curve[p_ix],
               '_pf_area_rel_zero': simps(peak_curve),
               '_pf_area_rel_min': simps(peak_curve - np.min(peak_curve)),
               '_pf_rising_slope_avg': abs(curve[b_ix_l] - curve[p_ix]) / abs(p_ix - b_ix_l),
               '_pf_falling_slope_avg': abs(curve[b_ix_r] - curve[p_ix]) / abs(p_ix - b_ix_r),
               '_pf_duration_base': peak_curve.size,
               '_pf_peak_curve': peak_curve,
               '_pf_uuid': str(uuid.uuid4()),
               '_pf_p_ix': p_ix,
               '_pf_b_ix_l': b_ix_l,
               '_pf_b_ix_r': b_ix_r
               }

        return pd.Series(out)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 17 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from .column_widget import ColumnWidget


class DataFrameColumnsWidget(QtWidgets.QWidget):
    # signal_filter_requested = QtCore.pyqtSignal(dict)

    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self, parent)
        # self.tab_name = tab_name

        self.verticalLayout = QtWidgets.QVBoxLayout(self)

        self.scrollArea = QtWidgets.QScrollArea(self)
        self.scrollArea.setWidgetResizable(True)

        self.scrollAreaWidgetContents = QtWidgets.QWidget()

        self.horizontalLayout = QtWidgets.QHBoxLayout(self.scrollAreaWidgetContents)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout.addWidget(self.scrollArea)

    def add_column(self, column_widget: ColumnWidget):
        # column_widget.signal_apply_clicked.connect(self.filter_requested)
        self.horizontalLayout.addWidget(column_widget)

    def add_horizontal_spacer(self):
        spacer = QtWidgets.QSpacerItem(1, 1, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacer)

    # @QtCore.pyqtSlot(dict)
    # def filter_requested(self, d):
    #     self.signal_filter_requested.emit(d)

if __name__ == '__main__':
    app = QtWidgets.QApplication([])
    w = DataFrameColumnsWidget(parent=None, tab_name='bah', filter_history='')

    cols = []
    for i in range(0,12):
        cols.append(ColumnWidget(parent=w, tab_name=w.tab_name, column_name='', column_type=str))
        w.add_column(cols[-1])
    w.resize(min(len(cols) * 215, 1920), 400)
    w.show()
    app.exec_()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'main_widget.ui'
#
# Created by: PyQt5 UI code generator 5.12
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWidget(object):
    def setupUi(self, MainWidget):
        MainWidget.setObjectName("MainWidget")
        MainWidget.resize(607, 500)
        MainWidget.setFloating(True)
        self.dockWidgetContents = QtWidgets.QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self.gridLayout = QtWidgets.QGridLayout(self.dockWidgetContents)
        self.gridLayout.setObjectName("gridLayout")
        self.comboBoxShowTimelineChoice = QtWidgets.QComboBox(self.dockWidgetContents)
        self.comboBoxShowTimelineChoice.setObjectName("comboBoxShowTimelineChoice")
        self.gridLayout.addWidget(self.comboBoxShowTimelineChoice, 0, 2, 1, 2)
        spacerItem = QtWidgets.QSpacerItem(242, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 2, 3, 1, 1)
        self.label = QtWidgets.QLabel(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 2)
        self.tabWidget = QtWidgets.QTabWidget(self.dockWidgetContents)
        self.tabWidget.setObjectName("tabWidget")
        self.gridLayout.addWidget(self.tabWidget, 1, 0, 1, 5)
        self.btnSetAllMaps = QtWidgets.QPushButton(self.dockWidgetContents)
        self.btnSetAllMaps.setObjectName("btnSetAllMaps")
        self.gridLayout.addWidget(self.btnSetAllMaps, 2, 4, 1, 1)
        MainWidget.setWidget(self.dockWidgetContents)

        self.retranslateUi(MainWidget)
        self.tabWidget.setCurrentIndex(-1)
        QtCore.QMetaObject.connectSlotsByName(MainWidget)

    def retranslateUi(self, MainWidget):
        _translate = QtCore.QCoreApplication.translate
        MainWidget.setWindowTitle(_translate("MainWidget", "Stimulus Mapping"))
        self.label.setText(_translate("MainWidget", "Show on timeline: "))
        self.btnSetAllMaps.setText(_translate("MainWidget", "Set all maps"))





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#@author: kushal

#Chatzigeorgiou Group
#Sars International Centre for Marine Molecular Biology

#GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007


import os
from stat import S_IEXEC
from time import time
from datetime import datetime
from . import get_sys_config
from typing import *
import numpy as np
import h5py
import json
import pandas as pd
from warnings import warn
import traceback
from graphviz import Digraph
from .configuration import IS_WINDOWS



def make_workdir(prefix: str = '') -> str:
    """
    Make a workdir within the mesmerize_tmp directory of the workdir specified in the configuration
    The name of the created workdir is the date & time of its creation. You can add a prefix to this name.

    :param prefix: Prefix for the workdir namec

    :return:    full workdir path
    :rtype:     str
    """

    main_workdir = get_sys_config()['_MESMERIZE_WORKDIR']

    if main_workdir == '':
        raise ValueError('You have not set the working directory in the System Configuration')
    if not os.access(main_workdir, os.W_OK):
        raise PermissionError(f'You do not have write permissions for the chosen work folder:\n{main_workdir}')

    date = datetime.fromtimestamp(time())
    dirname = f'{prefix}_{date.strftime("%Y%m%d")}_{date.strftime("%H%M%S")}_r{np.random.randint(0, 1000)}'
    workdir = os.path.join(main_workdir, 'mesmerize_tmp', dirname)
    os.makedirs(workdir)

    return workdir


# def make_runfile(*args, **kwargs) -> str:
#     if not IS_WINDOWS:
#         return _make_runfile_posix(*args, **kwargs)
#
#     elif IS_WINDOWS:
#         return _make_runfile_windows(*args, **kwargs)
#
#
# def _make_runfile_windows(module_path: str, workdir: str, args_str: str = None, filename: str = None,
#                  pre_run: str = '', post_run: str = '') -> str:
#
#     if filename is None:
#         sh_file = os.path.join(workdir, 'run')
#     else:
#         sh_file = os.path.join(workdir, filename)
#
#     sys_cfg = get_sys_config()



def make_runfile(module_path: str, savedir: str, args_str: Optional[str] = None, filename: Optional[str] = None,
                 pre_run: Optional[str] = None, post_run: Optional[str] = None) -> str:
    """
    Make an executable bash script. Used for running python scripts in external processes.

    :param module_path: absolute module path
    :type module_path:  str

    :param args_str:    str of args that is directly passed with the python command in the bash script
    :type args_str:     str

    :param savedir:     working directory
    :type savedir:      Optional[str]

    :param filename:    optional, specific filename for the script
    :type filename:     Optional[str]

    :param pre_run:     optional, str to run before module is ran
    :type pre_run:      Optional[str]

    :param post_run:    optional, str to run after module has run
    :type post_run:     Optional[str]

    :return: path to the shell script that can be run
    :rtype:  str
    """

    if filename is None:
        if IS_WINDOWS:
            sh_file = os.path.join(savedir, 'run.ps1')
        else:
            sh_file = os.path.join(savedir, 'run.sh')
    else:
        sh_file = os.path.join(savedir, filename)

    sys_cfg = get_sys_config()

    if pre_run is None:
        pre_run = ''
    if post_run is None:
        post_run = ''

    n_threads = sys_cfg['_MESMERIZE_N_THREADS']
    use_cuda = sys_cfg['_MESMERIZE_USE_CUDA']
    python_call = sys_cfg['_MESMERIZE_PYTHON_CALL']

    cmd_prefix = sys_cfg['_MESMERIZE_PREFIX_COMMANDS']

    if not os.path.isdir(savedir):
        try:
            os.makedirs(savedir)
        except PermissionError:
            raise PermissionError('You do not have permission to write to the chosen working directory.')

    elif os.path.isfile(savedir):
        raise FileExistsError("Choose a different working dir path")

    if args_str is None:
        args_str = ''

    if not IS_WINDOWS:
        to_write = '\n'.join(['#!/bin/bash',
                              f'{cmd_prefix}',
                              f'export _MESMERIZE_N_THREADS={n_threads}',
                              f'export _MESMERIZE_USE_CUDA={use_cuda}',
                              f'{pre_run}',
                              f'{python_call} {module_path} {args_str}',
                              f'{post_run}'])
    else:
        to_write = os.linesep.join(
            [
                f'{cmd_prefix}',
                f'$env:_MESMERIZE_N_THREADS="{n_threads}"',
                f'$env:_MESMERIZE_USE_CUDA="{use_cuda}"',
                f'{pre_run}',
                f'{python_call} {module_path} {args_str}',
                f'{post_run}'
            ]
        )

    with open(sh_file, 'w') as sf:
        sf.write(to_write)

    st = os.stat(sh_file)
    os.chmod(sh_file, st.st_mode | S_IEXEC)

    return sh_file


class HdfTools:
    """Functions for saving and loading HDF5 data"""

    @staticmethod
    def save_dataframe(path: str, dataframe: pd.DataFrame, metadata: Optional[dict] = None,
                       metadata_method: str = 'json', raise_meta_fail: bool = True):
        """
        Save DataFrame to hdf5 file along with a meta data dict.

        Meta data dict can either be serialized with json and stored as a str in the hdf5 file, or recursively saved
        into hdf5 groups if the dict contains types that hdf5 can deal with. Experiment with both methods and see what works best

        Currently the hdf5 method can work with these types: [str, bytes, int, float, np.int, np.int8, np.int16,
        np.int32, np.int64, np.float, np.float16, np.float32, np.float64, np.float128, np.complex].

        If it encounters an object that is not of these types it will store whatever that object's __str__() method
        returns if on_meta_fail is False, else it will raise an exception.

        :param path:            path to save the file to
        :type path:             str

        :param dataframe:       DataFrame to save in the hdf5 file
        :type dataframe:        pd.DataFrame

        :param metadata:        Any associated meta data to store along with the DataFrame in the hdf5 file
        :type metadata:         Optional[dict]

        :param metadata_method: method for storing the metadata dict, either 'json' or 'recursive'
        :type metadata_method:  str

        :param raise_meta_fail: raise an exception if recursive metadata saving encounters an unsupported object
                                If false, it will save the unsupported object's __str__() return value
        :type raise_meta_fail:  bool
        """

        if os.path.isfile(path):
            raise FileExistsError

        f = h5py.File(path, mode='w')

        f.create_group('DATAFRAME')

        if metadata is not None:
            mg = f.create_group('META')
            mg.attrs['method'] = metadata_method

            if metadata_method == 'json':
                bad_keys = []
                for k in metadata.keys():
                    try:
                        mg.create_dataset(k, data=json.dumps(metadata[k]))
                    except TypeError as e:
                        bad_keys.append(k + ': ' + str(e))

                if len(bad_keys) > 0:
                    bad_keys = '\n'.join(bad_keys)
                    raise TypeError(f"The following meta data keys are not JSON serializable\n{bad_keys}")

            elif metadata_method == 'recursive':
                HdfTools._dicts_to_group(h5file=f, path='META/', d=metadata, raise_meta_fail=raise_meta_fail)

        f.close()

        dataframe.to_hdf(path, key='DATAFRAME', mode='r+')

    @staticmethod
    def load_dataframe(filepath: str) -> Tuple[pd.DataFrame, Union[dict, None]]:
        """
        Load a DataFrame along with meta data that were saved using ``HdfTools.save_dataframe``

        :param filepath: file path to the hdf5 file
        :type filepath:  str

        :return: tuple, (DataFrame, meta data dict if present else None)
        :rtype: Tuple[pd.DataFrame, Union[dict, None]]
        """

        with h5py.File(filepath, 'r') as f:
            if 'META' in f.keys():

                if f['META'].attrs['method'] == 'json':
                    ks = f['META'].keys()
                    metadata = dict.fromkeys(ks)
                    for k in ks:
                        metadata[k] = json.loads(f['META'][k][()])

                elif f['META'].attrs['method'] == 'recursive':
                    metadata = HdfTools._dicts_from_group(f, 'META/')

            else:
                metadata = None
        df = pd.read_hdf(filepath, key='DATAFRAME', mode='r')

        return (df, metadata)

    @staticmethod
    def save_dict(d: dict, filename: str, group: str, raise_type_fail=True):
        """
        Recursively save a dict to an hdf5 group.

        :param d:        dict to save
        :type d:         dict

        :param filename: filename
        :type filename:  str

        :param group:    group name to save the dict to
        :type group:     str

        :param raise_type_fail: whether to raise if saving a piece of data fails
        :type raise_type_fail:  bool
        """

        if os.path.isfile(filename):
            raise FileExistsError

        with h5py.File(filename, 'w') as h5file:
            HdfTools._dicts_to_group(h5file, f'{group}/', d, raise_meta_fail=raise_type_fail)

    @staticmethod
    def _dicts_to_group(h5file: h5py.File, path: str, d: dict, raise_meta_fail: bool):
        for key, item in d.items():

            if isinstance(item, np.ndarray):

                if item.dtype == np.dtype('O'):
                    # see if h5py is ok with it
                    try:
                        h5file[path + key] = item
                        # h5file[path + key].attrs['dtype'] = item.dtype.str
                    except:
                        msg = f"numpy dtype 'O' for item: {item} not supported by HDF5\n{traceback.format_exc()}"

                        if raise_meta_fail:
                            raise TypeError(msg)
                        else:
                            h5file[path + key] = str(item)
                            warn(f"{msg}, storing whatever str(obj) returns.")

                # numpy array of unicode strings
                elif item.dtype.str.startswith('<U'):
                    h5file[path + key] = item.astype(h5py.special_dtype(vlen=str))
                    h5file[path + key].attrs['dtype'] = item.dtype.str  # h5py doesn't restore the right dtype for str types

                # other types
                else:
                    h5file[path + key] = item
                    # h5file[path + key].attrs['dtype'] = item.dtype.str

            # single pieces of data
            elif isinstance(item, (str, bytes, int, float, np.int, np.int8, np.int16, np.int32, np.int64, np.float,
                                   np.float16, np.float32, np.float64, np.float128, np.complex)):
                h5file[path + key] = item

            elif isinstance(item, dict):
                HdfTools._dicts_to_group(h5file, path + key + '/', item, raise_meta_fail)

            # last resort, try to convert this object to a dict and save its attributes
            elif hasattr(item, '__dict__'):
                HdfTools._dicts_to_group(h5file, path + key + '/', item.__dict__, raise_meta_fail)

            else:
                msg = f"{type(item)} for item: {item} not supported not supported by HDF5"

                if raise_meta_fail:
                    raise ValueError(msg)

                else:
                    h5file[path+key] = str(item)
                    warn(f"{msg}, storing whatever str(obj) returns.")

    @staticmethod
    def load_dict(filename: str, group: str) -> dict:
        """
        Recursively load a dict from an hdf5 group.

        :param filename: filename
        :type filename:  str

        :param group:    group name of the dict
        :type group:     str

        :return:         dict recursively loaded from the hdf5 group
        :rtype:          dict
        """

        with h5py.File(filename, 'r') as h5file:
            return HdfTools._dicts_from_group(h5file, f'{group}/')

    @staticmethod
    def _dicts_from_group(h5file: h5py.File, path: str) -> dict:
        ans = {}
        for key, item in h5file[path].items():
            if isinstance(item, h5py._hl.dataset.Dataset):
                if item.attrs.__contains__('dtype'):
                    ans[key] = item[()].astype(item.attrs['dtype'])
                else:
                    ans[key] = item[()]
            elif isinstance(item, h5py._hl.group.Group):
                ans[key] = HdfTools._dicts_from_group(h5file, path + key + '/')
        return ans


def draw_graph(l: List[dict], filename: Optional[str] = None, view: bool = False) -> str:
    """
    Draw a graph from a list of dicts.

    :param l: list of dicts
    :type l: List[dict]

    :param filename: full path for storing the draw graph pdf file
    :type filename: Optional[str]

    :param view: view the graph in the system's default pdf reader after it is rendered
    :type view: Optional[bool]

    :return: full path to the graph pdf file
    :rtype: str
    """
    if filename is None:
        workdir = make_workdir(prefix='graph_')
        filename = os.path.join(workdir, 'graph')

    g = Digraph('G', filename=filename, format='pdf', node_attr={'shape': 'record'})

    o_encounters = []
    o_n_l = []

    for op in l:
        o = list(op.keys())[0]
        params = op[o]
        s = []
        for k in params.keys():
            p = str(params[k]).replace("{", "\\n   --\> ").replace("}", "").replace("|", "\|")  # .replace("[", " ").replace("(", " ").replace(")", " ")
            s.append(f"{k}: {p}")

        o_n = f"{o}.{o_encounters.count(o)}"
        o_n_l.append(o_n)
        s = '{ ' + o_n + ' | ' + '\\n'.join(s) + ' }'

        g.node(o_n, s)
        o_encounters.append(o)

    es = []
    for i in range(1, len(o_n_l)):
        es.append((o_n_l[i - 1], o_n_l[i]))

    g.edges(es)
    g.render(view=view)

    return filename



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'CurvePlotWindow_template.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1209, 696)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1209, 23))
        self.menubar.setObjectName("menubar")
        self.menuView = QtWidgets.QMenu(self.menubar)
        self.menuView.setObjectName("menuView")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.dockFcWidget = QtWidgets.QDockWidget(MainWindow)
        self.dockFcWidget.setMinimumSize(QtCore.QSize(264, 36))
        self.dockFcWidget.setFloating(False)
        self.dockFcWidget.setFeatures(QtWidgets.QDockWidget.DockWidgetFloatable|QtWidgets.QDockWidget.DockWidgetMovable)
        self.dockFcWidget.setObjectName("dockFcWidget")
        self.dockWidgetContents = QtWidgets.QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self.dockFcWidget.setWidget(self.dockWidgetContents)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(2), self.dockFcWidget)
        self.dockConsole = QtWidgets.QDockWidget(MainWindow)
        self.dockConsole.setFeatures(QtWidgets.QDockWidget.DockWidgetFloatable|QtWidgets.QDockWidget.DockWidgetMovable)
        self.dockConsole.setObjectName("dockConsole")
        self.dockWidgetContents_2 = QtWidgets.QWidget()
        self.dockWidgetContents_2.setObjectName("dockWidgetContents_2")
        self.dockConsole.setWidget(self.dockWidgetContents_2)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(8), self.dockConsole)
        self.actionConsole = QtWidgets.QAction(MainWindow)
        self.actionConsole.setCheckable(True)
        self.actionConsole.setObjectName("actionConsole")
        self.actionControls_Widgets = QtWidgets.QAction(MainWindow)
        self.actionControls_Widgets.setCheckable(True)
        self.actionControls_Widgets.setChecked(True)
        self.actionControls_Widgets.setObjectName("actionControls_Widgets")
        self.menuView.addAction(self.actionConsole)
        self.menuView.addAction(self.actionControls_Widgets)
        self.menubar.addAction(self.menuView.menuAction())

        self.retranslateUi(MainWindow)
        self.actionConsole.toggled['bool'].connect(self.dockConsole.setVisible)
        self.actionControls_Widgets.toggled['bool'].connect(self.dockFcWidget.setVisible)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.menuView.setTitle(_translate("MainWindow", "View"))
        self.dockFcWidget.setWindowTitle(_translate("MainWindow", "Control Widgets"))
        self.dockConsole.setWindowTitle(_translate("MainWindow", "Console"))
        self.actionConsole.setText(_translate("MainWindow", "Console"))
        self.actionControls_Widgets.setText(_translate("MainWindow", "Controls Widgets"))

